(define-module (alterator vm profile)
  :use-module (srfi srfi-1)
  :use-module (srfi srfi-2)
  :use-module (srfi srfi-11)
  :use-module (srfi srfi-13)
  :use-module (alterator vm defs)
  :use-module (alterator vm utils)
  :export (create-profiles))

(define vm-profile-custom
  '(none (title . "Custom") (action . none)))

(define vm-profile-file "/tmp/metadata/vm-profile.scm")
(define vm-profile-file-bundled "/var/cache/alterator/vm-profile.scm")

(define (create-profiles)
  (let ((ext (or (and (access? vm-profile-file R_OK)
                      (map create-profile
                           (with-input-from-file vm-profile-file read)))
                 (and (access? vm-profile-file-bundled R_OK)
                      (map create-profile
                           (with-input-from-file vm-profile-file-bundled read)))
                 '())))
    (append ext (list (create-profile vm-profile-custom)))))

(define (profile-size profile)
  (if (eqv? (assv-ref  profile 'action) 'none) ""
      (size->string
       (apply + (map pr-min-size
                     (map (lambda (v) (assv-ref v 'size))
                          (map cdr (assv-ref profile 'actiondata))))))))

(define (create-profile profile)
  (let ((name (symbol->string (car profile)))
        (profile (cdr profile)))
    (lambda (command junk . args)
      (case command
        ((name) name)
        ((value) (lambda () profile))
        ((tree) (list name))
        ((title) `(quote ,(assv-ref profile 'title)))
        ((size) `(quote ,(profile-size profile)))
        ((apply)
         (format #t "APPLYING PROFILE: ~S ~S\n" name args)
         (and (assv-ref args 'clearall) (pr-clearall (assv-ref args 'commit)))
         (with-fluids
          ((pr-exclude (assv-ref args 'exclude)))
          (case (assv-ref profile 'action)
            ((trivial) (pr-trivial (assv-ref args 'commit)
                                   (assv-ref profile 'actiondata)))
            ((none) '(quote nexthop /vm/table))
            (else '(quote #f "Unknown profile type")))))))))

;---------------------------------------------------------------
(define pr-exclude (make-fluid))

(define (pr-clearall commit)
  (let ((kill (lambda (x) (false-if-exception (x 'destroy)))))
    (let ((killall (lambda (x) (for-each kill (apply evms x)))))
      (for-each killall
                '((volume-list)
                  (object-list #:type region #:data-type data)
                  (container-list)
                  (object-list #:type region #:data-type data)
                  (object-list #:type segment #:data-type data)))))
  (for-each
   (lambda (x) (and (x 'unassign?) (x 'unassign)))
   (evms 'object-list #:type 'disk))
  (and commit (evms 'commit)))

(define (pr-trivial commit data)
  (format #t "TRIVIAL: commit: ~S data: ~S\n" commit data)
  (if (every pr-place data)
      (if commit '(quote #t)
          '(quote nexthop /vm/table))
      '(quote #f "")))

(define (pr-place args)
  (let ((name (car args))
        (args (cdr args)))
    (format #t "PLACE: ~S ~S\n" name args)
    (pr-make-volume
     name (assv-ref args 'fsim)
     (pr-place-one name (assv-ref args 'size)
                   (assv-ref args 'methods)))))

(define (pr-make-volume name fsim object)
  (and object
       (format #t "MAKE-VOLUME: ~S ~S ~S\n" name fsim (obj->name (car object)))
       (and-let*
        ((vol (car ((evms 'create-compat (car object)) 'apply))))
        (format #t "MAKE-VOLUME: volume: ~S\n" (obj->name vol))
        (and fsim                       ;PReP doesn't have fsim
             (let ((task (evms 'mkfs (evms 'plugin #:name fsim))))
               (set! (task 'selected) (list vol))
               (task 'apply))
             (format #t "MAKE-VOLUME: fsim: ~S\n" (obj->plugin->name vol))
             ;; do not assign swap mntpoint, will be done later
             (or (equal? (obj->plugin->name vol) "SWAPFS")
                 (let ((task (evms 'setmntent vol)))
                   (set! (task 'options) `((mntpoint . ,name)))
                   (task 'apply))))
        (format #t "MAKE-VOLUME: finished\n")
        vol)))

(define (pr-place-one name size methods)
  (format #t "PLACE-ONE: ~S ~S ~S\n" name size methods)
  (if (null? methods) #f
      (or (case (car methods)
            ((plain) (pr-place-plain name size))
            ((raid+lvm) (format #t "raid+lvm\n") #f)
            ((raid) (pr-place-raid name size))
            ((lvm) (format #t "lvm\n") #f)
            (else (format #t "wtf?\n") #f))
          (pr-place-one name size (cdr methods)))))

(define (pr-place-size< placea placeb)
  ;; TODO complex criteria
  (let-values (((obja junka junktooa) placea)
               ((objb junkb junktoob) placeb))
    (< (obj->size obja) (obj->size objb))))

;---------------------------------------------------------------
(define (pr-place-plain name size)
  (let loop ((places (pr-list-places-plain size)))
    (and (not (null? places))
         (or (pr-place-plain-one (car places)
                                 (cond ((string=? name "prep") 65)
                                       ((string=? name "swap") 130)
                                       (#t 131))
                                 size)
             (loop (cdr places))))))

(define (pr-list-places-plain size)
  (sort
   (append-map (lambda(dsk) (pr-disk->freespace dsk size))
               (pr-candidates))
   pr-place-size<))

(define (pr-place-plain-one place type size)
  (false-if-exception
   (let-values (((seg splittable primary) place))
     (let ((task (evms 'create (evms 'plugin #:name "DosSegMgr"))))
       (format #t "MAKE-PARTITION: selected: ~S acceptable: ~S\n"
               (obj->name seg) (map obj->name (task 'acceptable)))
       (set! (task 'selected) (list seg))
       (and (number? (pr-max-size size))
            (set! (task 'options) `(("Size" . ,(pr-max-size size)))))
       (and (eqv? type 65)
            (set! (task 'options) `(("Bootable" . #t))))
       (set! (task 'options)
             `(("Primary" . ,primary)
               ("TypeByNumber" . ,type)))
       (task 'apply)))))

;---------------------------------------------------------------
(define (pr-place-raid name size)
  (let ((disks (pr-candidates)))
    (and (< 1 (length disks))
         ;; TODO / on raid5, check for separate boot
         (cond ((string=? name "/")
                (pr-raid-place 253 size 1 disks))
               (#t
                (or (pr-raid-place 131 size 2 disks)
                    (pr-raid-place 131 size 1 disks)))))))

(define (pr-raid-place type size minp disks)
  (let ((ndisks (length disks)))
    (and (< minp ndisks)
         (let* ((psize (pr-raid-scale-size size minp ndisks))
                (nplaces (pr-raid-list-places psize disks)))
           (if (any (lambda(p) (null? (cdr p))) nplaces)
               ;; filter out disk(s) without free space for raid part
               (pr-raid-place
                size type minp
                (filter-map (lambda(p) (and (not (null? (cdr p))) (car p))) nplaces))
               ;; okay, there are free spaces on remaining disks
               (let ((objs (append-map
                            (lambda(p) (pr-raid-place-part type psize p))
                            (map cdr nplaces))))
                 (if (every values objs)
                     (let ((task (evms 'create (pr-raid-plugin minp))))
                       (set! (task 'selected) objs)
                       (task 'apply))
                     (let ()
                       (for-each (lambda(obj) (obj 'delete))
                                 (filter values objs)
                                 #f)))))))))

(define (pr-raid-place-part type size places)
  (and (not (null? places))
       (or (pr-place-plain-one (car places) type size)
           (pr-raid-place-part type size (cdr places)))))

(define (pr-raid-list-places size disks)
  (map
   (lambda(dsk)
     (cons dsk (sort (pr-disk->freespace dsk size) pr-place-size<)))
   disks))

(define (pr-raid-scale-size size minp n)
  (if (= 1 minp) size                   ;raid1
      (pr-scale-size (lambda(s) (round-integer (/ s (- n 1)))) size)))

(define (pr-raid-plugin minp)
  (evms 'plugin #:name
        (if (= 1 minp)
            "MDRaid1RegMgr"
            "MDRaid5RegMgr")))

;---------------------------------------------------------------
(define (pr-disk->freespace disk size)
  (let ((parents (obj->parents disk)))
    (format #t "DISK->FREESPACE: ~S parents: ~S\n" (obj->name disk) (map obj->name parents))
    (cond  ((null? parents)        ; no partition table yet
            (format #t "DISK->FREESPACE: no pt on ~S, creating\n" (obj->name disk))
            (pr-create-pt disk)
            (pr-disk->freespace disk size))
           ((not (pr-disk-dos? (car parents)))
            (format #t "DISK->FREESPACE: not a dos pt, skipping\n")
            '())
           (#t
            (let ((numslots (apply pr-numslots (map obj->name (cons disk (filter segment-data? parents))))))
              (format #t "DISK->FREESPACE: SLOTS: ~S\n" numslots)
              (fold
               (lambda (seg accum)
                 (if (and (segment-free? seg) (pr-size< size (obj->size seg)))
                     (let-values (((before after) (partition (lambda(segn) (pr-seg< segn seg)) parents)))
                       (format #t "DISK->FREESPACE: ~S X ~S\n"
                               (map obj->name before) (map obj->name (cdr after)))
                       (cond ((< 3 numslots)
                              ;; empty disk: splittable, primary
                              (cons (values seg #t #t) accum))
                             ((not (null? (filter segment-meta? after)))
                              ;; freespace, then ext'd: maybe-splittable, primary
                              (cons (values seg (< 2 numslots) #t) accum))
                             ((< 1 (length (filter segment-meta? before)))
                              ;; ext'd, then free
                              (if (segment-meta? (cadr (reverse before)))
                                  ;; no pri after ext'd: splittable, logical
                                  (cons (values seg #t #f) accum)
                                  ;; pri after ext'd: maybe-splittable, primary
                                  (cons (values seg (< 2 numslots) #f) accum)))
                             (#t
                              ;; primary only
                              (cons (values seg (< 2 numslots) (< 2 numslots)) accum))))
                     accum))
               '() parents))))))

(define (pr-candidates)
  (filter pr-wholedisk?
          (filter pr-candidate?
                  (evms 'object-list #:type 'disk))))

(define (pr-candidate? disk)
  (not (member (obj->name disk)
               (fluid-ref pr-exclude))))

(define (pr-wholedisk? disk)
  (and (pr-purge-fake-volume disk) disk))

(define (pr-purge-fake-volume disk)
  (not (and-let*
        ((vol (assv-ref (disk) 'volume)))
        (or (assv-ref (vol) 'fsim)
            (begin (vol 'delete) #f)))))

(define (pr-disk-dos? obj)
  (string=? "DosSegMgr" (obj->plugin->name obj)))

(define (pr-numslots diskname . segnames)
  (format #t "PR-NUMSLOTS: ~S ~S\n" diskname segnames)
  (let ((pslots (map (lambda(s) (string-append diskname s)) '("1" "2" "3" "4"))))
    (let-values (((pri log) (partition (lambda(x) (member x pslots)) segnames)))
      (- 4 (length pri) (if (null? log) 0 1)))))

(define (pr-create-pt disk)
  (let ((task (evms 'assign (evms 'plugin #:name "DosSegMgr"))))
    (set! (task 'selected) (list disk))
    (task 'apply)))

(define (pr-seg< sega segb)
  (< (assv-ref (sega) 'start)
     (assv-ref (segb) 'start)))

(define (pr-size< size num)
  (< (pr-min-size size) num))

(define (pr-scale-size fn size)
  (if (pair? size)
      (cons (fn (car size))
            (let ((max (cdr size)))
              (if (number? max)
                  (fn max)
                  max)))
      (fn size)))

(define (pr-min-size size)
  (if (pair? size) (car size) size))

(define (pr-max-size size)
  (if (pair? size) (cdr size) size))
