(define-module (alterator vm task)
  :use-module (srfi srfi-1)
  :use-module (srfi srfi-2)
  :use-module (srfi srfi-11)
  :use-module (srfi srfi-13)
  :use-module (alterator vm defs)
  :use-module (alterator vm utils)
  :use-module (alterator vm option)
  :use-module (alterator vm tree))

;---------------------------------------------------------------
(define fsim-alist
  '(("Ext2/3" "131" "142" "253")
    ("XFS" "131" "142" "253")
    ("JFS" "131" "142" "253")
    ("ReiserFS" "131" "142" "253")
    ("SWAPFS" "130")
    ("FAT16" "4" "6" "14" "20" "22")
    ("FAT32" "11" "12" "27" "28")
    ("NTFS" "7" "23")))

(define (evms-fsims args)
  (let ((type (assv-ref args 'type-by-number)))
    (if type
        (let ((fsims (filter-map
                      (lambda (elt)
                        (and (member type (cdr elt))
                             (evms 'plugin #:name (car elt))))
                      fsim-alist)))
          (if (= (length fsims) 1)
              (car fsims)
              fsims))
        (map
         (lambda (name) (evms 'plugin #:name name))
         (map car fsim-alist)))))

(define (create? op plugin)
  (and-let*
   ((task (false-if-exception (evms op plugin)))
    (acceptable (false-if-exception (not (null? (task 'acceptable))))))
   (task 'destroy)
   #t))

(define (create-segment? obj)
  (and-let*
   ((task (false-if-exception (evms 'create (evms 'plugin #:name "DosSegMgr"))))
    (acceptable (false-if-exception (task 'acceptable))))
   (task 'destroy)
   (case (obj->type obj)
     ((disk) (and (any (lambda (x) (memv x acceptable)) (obj->parents obj)) #t))
     (else (and (memv obj acceptable) #t)))))

(define (mkfs? obj)
  (any (lambda (fsim) (false-if-exception (obj 'mkfs? fsim)))
       (map
        (lambda (name) (evms 'plugin #:name name))
        (map car fsim-alist))))

(define (setmntent? vol)
  (and (assv-ref (vol) 'fsim) #t))

(set! operations
      (lambda (thing)
        (case (obj->type thing)
          ((plugin)
           (let ((name (obj->name thing)))
             (cond
              ((equal? name "MDRaid1RegMgr")
               `((op-raid-create . ,(create? 'create thing))))
              ((equal? name "LVM2")
               `((op-vg-create . ,(create? 'create-container thing))))
              (#t '()))))
          ((disk)
           (if (null? (obj->parents thing))
               ;; no segmgr yet, we can assign segmgr, or use whole disk
               `((op-pt-create . ,(thing 'assign?))
                 (op-volume-create . ,(thing 'create-compat?)))
               ;; OR, create segment or remove segmgr
               `((op-segment-create . ,(create-segment? thing))
                 (op-pt-destroy . ,(thing 'unassign?)))))
          ((segment)
           ;; allow create segment on explicitly selected freespace
           (case (obj->dtype thing)
             ((free)
              `((op-segment-create . ,(create-segment? thing))))
             (else
              `((op-segment-expand . ,(thing 'expand?))
                (op-segment-shrink . ,(thing 'shrink?))
                (op-segment-destroy . ,(thing 'destroy?))
                (op-volume-create . ,(thing 'create-compat?))))))
          ((region)
           (let ((plugin (obj->plugin->name thing)))
             (cond
              ((member plugin '("MDRaid0RegMgr" "MDRaid1RegMgr" "MDRaid5RegMgr"))
               `((op-volume-create . ,(thing 'create-compat?))
                 (op-raid-destroy . ,(thing 'destroy?))))
              ((equal? plugin "LVM2")
               `((op-lv-create . #t)          ; FIXME pred
                 (op-volume-create . ,(thing 'create-compat?))
                 (op-lv-expand . ,(thing 'expand?))
                 (op-lv-shrink . ,(thing 'shrink?))
                 (op-lv-destroy . ,(thing 'destroy?))))
              (#t '()))))
          ((container)
           (let ((plugin (obj->plugin->name thing)))
             (cond
              ((equal? plugin "LVM2")
               `((op-vg-expand . ,(thing 'expand?))
                 (op-vg-shrink . ,(thing 'shrink?))
                 (op-vg-destroy . ,(thing 'delete?))))
              (#t '()))))
          ((volume)
           `((op-volume-mkfs . ,(mkfs? thing))
             (op-volume-fsck . ,(thing 'fsck?))
             (op-volume-unmkfs . ,(thing 'unmkfs?))
             (op-volume-expand . ,(thing 'expand?))
             (op-volume-shrink . ,(thing 'shrink?))
             (op-volume-mount . ,(thing 'mount?))
             (op-volume-umount . ,(thing 'umount?))
             (op-volume-remount . ,(thing 'remount?))
             (op-volume-setmntent . ,(setmntent? thing))
             (op-volume-destroy . ,(thing 'destroy?))))
          (else '()))))

(define task-id
  (let ((counter 0))
    (lambda () (set! counter (1+ counter)) counter)))

(define (task-path task)
  (string-append "/evms/tasks/" (name-of task)))

(define (create-task obj . args)
  (format #t "CREATE-TASK: ~S ~S\n" obj args)
  (let ((start (assv-ref args 'start))
        (verbose (assv-ref args 'verbose))
        (op (assv-ref args 'operation))
        (args (alist-delete 'start (alist-delete 'verbose (alist-delete 'operation args)))))
    (let-values (((op obj actual-op actual-obj next-op next-obj initially-verbose) (mangle-task-args op obj args)))
      (cond
       ((procedure? actual-obj)         ; will return ordinary task
        (create-task-node op obj
                          next-op next-obj
                          actual-op actual-obj
                          (if start initially-verbose verbose)
                          args))
       ((list? actual-obj)              ; will return list of tasks
        (let* ((verbose (if start initially-verbose verbose))
               (tasks (filter-map (lambda (object)
                                    (create-task-node op obj
                                                      next-op next-obj
                                                      actual-op object
                                                      verbose
                                                      args))
                                  actual-obj)))
          (if (null? tasks) #f tasks)))))))

(define (mangle-task-args op obj args)
  (case op
    ((op-pt-create)
     ;; (values op obj 'assign (evms 'plugin #:name "DosSegMgr") 'op-segment-create obj #t))
     ;; short-circuit to op-segment-create
     (let ((task (evms 'assign (evms 'plugin #:name "DosSegMgr"))))
       (set! (task 'selected) (list obj))
       (task 'apply))
     (let ((obj (cadr (assv-ref (obj) 'parents))))  ; second seg maps to freespace
       (format #t "OP-PT-CREATE: ~S\n" (obj))
       (values 'op-segment-create obj 'create (evms 'plugin #:name "DosSegMgr") 'op-volume-create #f #t)))
    ((op-pt-destroy)
     (values op obj 'unassign obj #f #f #t))
    ((op-segment-create)
     (values op obj 'create (evms 'plugin #:name "DosSegMgr") 'op-volume-create #f #t))
    ((op-raid-create)
     (values op obj 'create obj 'op-volume-create #f #t))
    ((op-vg-create)
     (values op obj 'create-container obj 'op-lv-create #f #t))
    ((op-lv-create)
     (values op obj 'create (obj->plugin obj) 'op-volume-create #f #t))
    ((op-segment-expand op-vg-expand op-lv-expand)
     (values op obj 'expand obj #f #f #t))
    ((op-segment-shrink op-vg-shrink op-lv-shrink)
     (values op obj 'shrink obj #f #f #t))
    ((op-volume-create)
     (values op obj 'create-compat obj 'op-volume-mkfs #f #t))
    ((op-volume-expand)
     (and (equal? (obj->plugin->name obj) "NTFS")
          (fluid-set! current-confirmation
                      "Note, that resizing Vista partition may lead to Vista boot failure. Proceed?"))
     (values op obj 'expand (obj->obj obj) #f #f #t))
    ((op-volume-shrink)
     (and (equal? (obj->plugin->name obj) "NTFS")
          (fluid-set! current-confirmation
                      "Note, that resizing Vista partition may lead to Vista boot failure. Proceed?"))
     (values op obj 'shrink (obj->obj obj) #f #f #t))
    ((op-volume-mkfs)
     (if evms-prefix
         (values op obj 'mkfs (evms-fsims args) 'op-volume-setmntent obj #t)
         (values op obj 'mkfs (evms-fsims args) #f obj #t)))
    ((op-volume-fsck)
     (values op obj 'fsck obj #f #f #t))
    ((op-volume-unmkfs)
     (values op obj 'unmkfs obj #f #f #t))
    ((op-volume-mount)
     (values op obj 'mount obj #f #f #t))
    ((op-volume-setmntent)
     (values op obj 'setmntent obj #f #f #t))
    ((op-volume-umount)
     (values op obj 'umount obj #f #f #t))
    ((op-volume-remount)
     (values op obj 'remount obj #f #f #t))
    ((op-segment-delete op-vg-delete op-lv-delete op-raid-delete op-volume-delete)
     (values op obj 'delete obj #f #f #t))
    ((op-segment-destroy op-vg-destroy op-lv-destroy op-raid-destroy op-volume-destroy)
     (values op obj 'destroy obj #f #f #t))))

(define (select-one-object task obj)
  (set! (task 'selected) (list obj)) #t)

(define (select-one-acceptable task obj objs)
  (and (memv obj objs)
       (select-one-object task obj)))

(define (select-acceptable task objs)
  (and (not (null? objs))
       (false-if-exception (set! (task 'selected) objs))
       ;; avoid selection from one object
       (or (null? (cdr objs)) '())))

(define (select-task-objects task op obj)
  ;; return #t if no further selection required, '() otherwise
  ;; #f cancels task
  (and-let*
   ((acceptable (false-if-exception (task 'acceptable))))
   (case op
     ((op-pt-create) (select-one-object task obj))
     ((op-segment-create)
      (case (obj->type obj)
        ((disk)
         ;; select freespace on given disk
         (select-acceptable
          task (filter (lambda(o) (memv obj (obj->children o))) acceptable)))
        ((segment) (select-one-object task obj))))
     ((op-segment-expand op-segment-shrink)
      (select-acceptable task acceptable))
     ((op-vg-create op-vg-expand op-raid-create)
      (select-acceptable task acceptable))
     ((op-lv-create)
      (case (obj->type obj)
        ((container)
         ;; select freespace on given container
         (select-acceptable
          task (filter (lambda(o) (eqv? (obj->producer o) obj)) acceptable)))
        ;; FIXME noop right now, allow op on freespace probably
        (else (set! (task 'selected) acceptable) #t)))
     ((op-volume-create) (select-one-object task obj))
     ((op-volume-mkfs op-volume-fsck op-volume-mount op-volume-umount op-volume-unmkfs op-volume-setmntent)
      (select-one-acceptable task obj acceptable))
     ((op-pt-destroy op-segment-destroy op-vg-destroy op-lv-destroy op-raid-destroy op-volume-destroy)
      (select-one-object task obj))
     ((op-volume-expand op-volume-shrink)
      (if (segment? obj)
          (select-acceptable task acceptable)
          '()))
     (else ; free for all: op-vg-shrink
      '()))))

(define (create-task-proc op obj context)
  (and-let*
   ((task (false-if-exception (apply evms op obj context))))
   (set-procedure-property! task 'transform '())
   task))

(define (create-task-node op obj next-op next-obj mangled-op mangled-obj verbose args)
  (format #t "CREATE-TASK-NODE: ~S ~S ~S ~S ~S\n" op mangled-op (obj->name mangled-obj) verbose args)
  (and-let*
   ((task (create-task-proc mangled-op mangled-obj args))
    (selection (select-task-objects task op obj))
    (id (number->string (task-id))))
   (lambda (command junk . args)
     (case command
       ((name) id)
       ((value) task)
       ((verbose) `(quote ,verbose))
       ((action) (list op next-op))
       ((tree) (list id))
       ((read) `(title ,(assv-ref (mangled-obj) 'long-name) next-op ,next-op))
       ((objects) (if (null? selection) `(quote ,@(map obj->name (task 'acceptable))) '()))
       ((options) `(quote ,@(task-options op obj task)))
       ((set-objects!)
        (format #t "SET-OBJECTS! ~S\n" (assv-ref args 'objects))
        (let-values
            (((effect declined-objects)
              (set! (task 'selected)
                    (filter (lambda (obj) (member (obj->name obj) (assv-ref args 'objects)))
                            (task 'acceptable)))))
          (format #t "DECLINED: ~S\n" (map obj->name declined-objects))
          effect))
       ((set-options!)
        (let ((args (fold apply args (map cdr (procedure-property task 'transform)))))
          (format #t "SET-OPTIONS! ~S\n" args)
          (catch 'swig-system-error
            (lambda () `(quote #t ,@(set! (task 'options) args)))
            (lambda (key funcname msg . args) `(quote #f ,msg)))))
       ((apply)
        (format #t "TASK-APPLY: ~S\n" args)
        (let ((verbose (assv-ref args 'verbose))
              (next-hop (assv-ref args 'next-hop)))
          (catch 'swig-system-error ; FIXME return paths to newly created objs
            (lambda ()
              (let ((context (procedure-property task 'context))
                    (objs (task 'apply)))
                (format #t "TASK-NEXT(~S): ~S ~S ~S ~S\n" next-hop next-op next-obj objs context)
                (let ((next-obj (or next-obj (and (not (null? objs)) (car objs)))))
                  (if (and next-hop next-op next-obj)
                      (apply tasks 'new '() next-obj `(verbose . ,verbose) `(operation . ,next-op) (or context '()))
                      `(quote #t)))))
            (lambda (key funcname msg . args) `(quote #f ,msg)))))
       ((destroy)
        (catch 'swig-system-error
          (lambda () (task 'destroy) `(quote #t))
          (lambda (key funcname msg . args) `(quote #f ,msg))))))))

(set! tasks
      (let* ((task-alist '())
             (task-add! (lambda (t) (set! task-alist (acons (name-of t) t task-alist))))
             (task-remove! (lambda (n) (set! task-alist (alist-delete! n task-alist))))
             (node (create-node "tasks")))
        (lambda (command path . args)
          (if (null? path)
              (case command
                ((new)
                 (with-fluids
                  ((current-confirmation #f))
                  (and-let*
                   ((result (apply create-task args)))
                   (cond
                    ((procedure? result)     ; will return ordinary task
                     (task-add! result)
                     `(quote task ,(fluid-ref current-confirmation)
                             ,@(result 'action '()) ,(task-path result)))
                    ((list? result)          ; will return list of tasks
                     (for-each task-add! result)
                     `(quote choice ,(fluid-ref current-confirmation)
                             ,@((car result) 'action '()) ,@(map task-path result)))))))
                ((delete) '())
                ((tree)
                 (let ((children (append-map tree-of (map cdr task-alist))))
                   (if (null? children) (list (name-of node))  ;;'()
                       (list (name-of node) children))))
                ((list)
                 (map list (map car task-alist)))
                (else (apply node command '() args)))
              (apply (assoc-ref task-alist (car path))
                     command (cdr path) args)))))
