(define-module (alterator vm tree)
  :use-module (srfi srfi-1)
  :use-module (srfi srfi-2)
  :use-module (srfi srfi-13)
  :use-module (alterator str)
  :use-module (alterator vm defs)
  :use-module (alterator vm utils)
  :use-module (alterator vm profile)
  :export (create-node
           create-compound
           create-tree))

;---------------------------------------------------------------
(define tree-hide-empty-branches #f)

(define (tree-item name obj)
  (case (obj->type obj)
    ((disk) `(#((,name . "vm/disk.png") ,@(properties-for obj '(size #f #f #f)))))
    ((segment) `(#((,name . ,(if (eqv? (obj->dtype obj) 'free) "vm/unused.png" "vm/partition.png"))
                   ,@(properties-for obj '(size #f #f #f)))))
    ((region container) `(#((,name . ,(if (eqv? (obj->dtype obj) 'free) "vm/unused.png" "vm/partition.png"))
                            ,@(properties-for obj '(size #f #f #f)))))
    ((volume)
     `(#((,name . ,(if (assv-ref (obj) 'mountpoint) "vm/mounted.png" "vm/umounted.png"))
         ,@(properties-for obj (list 'size 'fsim (if evms-prefix 'mntpoint 'mountpoint) 'mntoptions)))))
    (else `(#(,name "" "")))))

(define (size->string size)
  (if (< size 20971520)
      (string-append (number->string (round-integer (/ size 2048))) " MB")
      (string-append (number->string (round-integer (/ size 2097152))) " GB")))

(define (properties-for obj props)
  (let ((prop-alist (append (obj) (evms 'options obj `(force-mkfs . ,props)))))
    (map
     (lambda (key)
       (let ((value (assv-ref prop-alist key)))
         (cond                        ; FIXME more serialization cases
          ((eqv? key 'fsim)
           (let ((fsim (assv-ref prop-alist 'fsim)))
             (if fsim
                 (cons (obj->name fsim)
                       (if (and (assv-ref prop-alist 'mntpoint)
                                (assv-ref prop-alist 'force-mkfs))
                           "vm/caution.png"
                           "theme:null"))
                 "")))
          ((eqv? key 'mntoptions)
           (or (and (assv-ref prop-alist 'mntpoint)
                    (assv-ref prop-alist 'mntoptions))
               ""))
          ((eqv? key 'size)
           (let ((all (or
                       ;; new fssize after expanding will be shown
                       ;; only after commit. that sucks, so, volsize.
                       ;;(and-let* ((sz (assv-ref prop-alist 'fssize)))
                       ;;          (and (< 0 sz) sz))
                       (assv-ref prop-alist 'volsize)
                       (assv-ref prop-alist 'size)))
                 (used (assv-ref prop-alist 'minfssize)))
             (if used
                 (string-append (size->string all)
                                " [" (size->string (- all used)) "]")
                 (size->string all))))
          ((eqv? key 'free)
           (if (eqv? (obj->type obj) 'disk)
               (size->string (apply + (map obj->size (filter segment-free? (obj->parents obj)))))
               ""))
          ((procedure? value)
           (obj->name value))
          ((not value) "")
          (#t (->string value)))))
     props)))

(define (create-node thing)
  (apply create-node-proc (normalize thing)))

(define (create-node-proc name first . rest) ; one or more objects of same type
  (let ((value (if (null? rest) first (cons first rest))))
    (lambda (command junk . args)
      (case command
        ((name) name)
        ((value) value)
        ((tree) (tree-item name value))
        ((task) (apply tasks 'new '() value '(start . #t) args))
        ((ops) `(quote ,@(operations first)))
        ((props) `(quote ,@(properties-for first (assv-ref args 'props))))))))

(define (mangle-name obj)
  (string-join (string-split (obj->name obj) #\/) "|" 'infix))

(define (truncate-name obj)
  (car (reverse (string-split (obj->name obj) #\|))))

(define (normalize thing)
  (cond ((string? thing)
         (list thing (lambda x `((name . ,thing)))))
        ((eqv? (obj->type thing) 'plugin)
         (mangle-name thing))            ; for Ext2/3
        ((memq (obj->type thing) '(disk segment region feature container volume))
         (list (case (obj->dtype thing)
                 ((free)
                  (string-append
                   "<unused" (string-filter (obj->name thing) char-numeric?) ">"))
                 (else (truncate-name thing)))
               thing))
        (#t thing)))

(define (sort-nodes parent-obj nodes)
  (cond
   ((eqv? (obj->type parent-obj) 'disk)
    (sort nodes (lambda (a b)
                  (< (obj->start (value-of a)) (obj->start (value-of b))))))
   (#t (sort nodes (lambda (a b) (string< (name-of a) (name-of b)))))))

(define (alist-merge = lst)
  (map
   (lambda (elt)
     (cons elt (append-map cdr (filter (lambda (p) (= (car p) elt)) lst))))
   (delete-duplicates (map car lst))))

(define (create-compound node . nodes)
  (let ((contents (map (lambda (x) (cons (name-of x) x)) nodes)))
    (lambda (command path . args)
      (if (null? path)
          (case command
            ((tree)
             (let ((parent (tree-item (name-of node) (value-of node)))
                   (children (append-map tree-of (sort-nodes (value-of node) nodes))))
               (if (null? children) (if tree-hide-empty-branches '() parent)
                   (append parent `(,children)))))
            ((list)
             (map list (map car contents)))
            (else (apply node command '() args)))
          (apply (assoc-ref contents (car path))
                 command (cdr path) args)))))

(define (apply-compound arglist)
  (apply create-compound (map create-node arglist)))

(define (create-subtree nodespec . limits)
  (apply create-compound (create-node nodespec)
         (map apply-compound
              (alist-merge
               eqv?
               (apply append-map object-tuple (apply zip limits))))))

(define (create-disk-subtree)
  (create-subtree
   `("Disks"
     ,(evms 'plugin #:name "DosSegMgr"))
   `(,segment? ,disk?) `(,disk? ,(lambda (x) #f))))
; (pretty-print ((create-disk-subtree) 'tree '()))
; (pretty-print ((create-disk-subtree) 'read '()))

(define (create-raid-subtree)
  (create-subtree
   `("RAID"
     ,(evms 'plugin #:name "MDRaid1RegMgr")
     ,(evms 'plugin #:name "MDRaid0RegMgr")
     ,(evms 'plugin #:name "MDRaid5RegMgr"))
   ;;`(,raid? ,(lambda (x) (or (segment? x) (disk? x))))
   `(,raid? ,(lambda (x) #f))))
; (pretty-print ((create-raid-subtree) 'tree '()))
; (pretty-print ((create-raid-subtree) 'read '()))
; (pretty-print ((create-raid-subtree) 'task '() '(operation . create)))

(define (create-lvm-subtree)
  (create-subtree
   `("LVM"
     ,(evms 'plugin #:name "LVM2"))
   `(,lvm? ,container?)))
; (pretty-print ((create-lvm-subtree) 'tree '()))
; (pretty-print ((create-lvm-subtree) 'read '()))
; (pretty-print ((create-lvm-subtree) 'task '() '(operation . create-container)))


(define (create-tree control-node)
  (create-compound
   (create-node "evms")
   (create-compound
    (create-node "tree")
    (create-disk-subtree)
    (create-raid-subtree)
    (create-lvm-subtree))
   (apply create-compound
          (create-node "storage")
          (map apply-compound
               `(("disks" . ,(evms 'object-list #:type 'disk))
                 ;;("segments" . ,(evms 'object-list #:type 'segment))
                 ;;("regions" . ,(evms 'object-list #:type 'region))
                 ;;("features" . ,(evms 'object-list #:type 'feature))
                 ;;("containers" . ,(evms 'container-list))
                 ("volumes" . ,(evms 'volume-list)))))
   tasks
   (apply create-compound (create-node "profiles") (create-profiles))
   control-node))

(define (big-enough? obj)               ; FIXME
  (and-let* ((size (obj->size obj)))
            (< 16384 size)))

(define (objects)
  (append
   (evms 'volume-list)
   (evms 'object-list #:flags '(top) #:data-type 'data)
   (filter big-enough? (evms 'object-list #:flags '(top) #:data-type 'free))))

(define (expand-obj obj pred)
  (letrec
      ((w (lambda (obj)
            (cond
             ((not obj) '())
             ((pred obj) (list obj))
             (#t (next obj)))))
       (next (lambda (obj)
               (case (obj->type obj)
                 ((volume)
                  (w (obj->obj obj)))
                 ((region)
                  (if (assv-ref (obj) 'producer)
                      (w (assv-ref (obj) 'producer))
                      (append-map w (assv-ref (obj) 'children))))
                 ((container)
                  (append-map w (assv-ref (obj) 'consumed)))
                 (else (append-map w (assv-ref (obj) 'children)))))))
    (w obj)))

(define (object-tuple node-pred leaf-pred)
  (map
   (lambda (obj)
     (if (eqv? node-pred raid?)         ; FIXME ugly
         (cons obj (expand-obj obj leaf-pred))
         (append (expand-obj obj leaf-pred) (list obj))))
   (filter node-pred (objects))))

;; (map obj->type (expand-obj (car (filter lvm? (objects))) container?))
;; (apply map obj->type (object-tuple segment? disk?))
;; (apply map obj->type (object-tuple raid? (lambda (o) (or (segment? o) (disk? o)))))
;; (apply map obj->type (object-tuple lvm? container?))
;; (map obj->type (objects))

